//----------------------------------------------------------
// rt_Skin.cpp
//
// Sample to demonstrate the effects of vertex blending
// with a variety of weight distribution functions and
// transformations.
//
// This file is a modified version of the VertexBlend
// sample in the SDK.
//
// Rich Thomson
// July 15th, 2001
//
#include "rt_Skin.h"    // application header

#include "rt/mesh.h"    // mesh_vertex_lock
#include "rt/mat.h"     // mat_trans, mat_scale, ...
#include "rt/misc.h"    // NUM_OF
#include "rt/states.h"  // s_rs, set_states

//----------------------------------------------------------
// BLENDVERTEX
//
// Vertex structure that includes a blend weight.
//
struct BLENDVERTEX          // Referenced in the shader as:
{
    D3DXVECTOR3 v;          // v0
    float       blend;      // v1.x
    D3DXVECTOR3 n;          // v3
    float       tu, tv;     // v7
};
const DWORD BLENDVERTEX_FVF =
    D3DFVF_XYZB1 | D3DFVF_NORMAL | D3DFVF_TEX1;

//----------------------------------------------------------
// ConfirmDevice()
//
// Validate both vertex blending and vertex shader caps
// and set available application options accordingly.
// Check that the device supports at least one of the
// two techniques used in this sample: either a vertex
// shader, or at least two blend matrices and a
// directional light.
//
HRESULT
CMyD3DApplication::ConfirmDevice(D3DCAPS8 *pCaps,
    DWORD dwBehavior, D3DFORMAT Format)
{
    if (dwBehavior & D3DCREATE_SOFTWARE_VERTEXPROCESSING)
    {
        // Software vertex processing always supports what
        // we need
        return S_OK;
    }

    if (pCaps->VertexShaderVersion >= D3DVS_VERSION(1,0))
    {
        // HW device supports a vertex shader
        return S_OK;
    }

    // Check for vertex blending with at least two matrices
    // (Software can always do up to 4 blend matrices)
    if (pCaps->MaxVertexBlendMatrices < 2)
    {
        return E_FAIL;
    }

    // check for directional lights
    if (!(pCaps->VertexProcessingCaps &
            D3DVTXPCAPS_DIRECTIONALLIGHTS))
    {
        return E_FAIL;
    }

    return S_OK;
}

//----------------------------------------------------------
// InitDeviceObjects()
//
// Load the mesh and adjust behavior of app to follow caps
// of the device.
//
HRESULT
CMyD3DApplication::InitDeviceObjects()
{
    // Init the font
    m_font->InitDeviceObjects(m_pd3dDevice);

    // Load an object to render
    if (FAILED(m_mesh->Create(m_pd3dDevice,
        _T("mslogo.x"))))
    {
        return D3DAPPERR_MEDIANOTFOUND;
    }

    // compute bounding box
    {
        ID3DXMesh *mesh = m_mesh->GetSysMemMesh();
        rt::mesh_vertex_lock<BLENDVERTEX> lock(mesh);
        THR(::D3DXComputeBoundingBox(lock.data(),
            mesh->GetNumVertices(), mesh->GetFVF(),
            &m_mesh_bounds.m_min, &m_mesh_bounds.m_max));
    }

    // adjust behavior to follow caps
    if ((m_dwCreateFlags &
            (D3DCREATE_HARDWARE_VERTEXPROCESSING |
             D3DCREATE_MIXED_VERTEXPROCESSING)) &&
        m_d3dCaps.VertexShaderVersion < D3DVS_VERSION(1,0))
    {
        // No vertex shaders
        m_use_vertex_shader = false;
        ::EnableMenuItem(::GetMenu(m_hWnd),
            IDM_USEVERTEXSHADER, MF_GRAYED);
    }
    else if (m_d3dCaps.MaxVertexBlendMatrices < 2)
    {
        // No blend matrices available
        m_use_vertex_shader = true;
        ::EnableMenuItem(::GetMenu(m_hWnd),
            IDM_USEVERTEXSHADER, MF_GRAYED);
    }
    else
    {
        // Both techniques available
        m_use_vertex_shader = false;
        ::EnableMenuItem(::GetMenu(m_hWnd),
            IDM_USEVERTEXSHADER, MF_ENABLED);
    }

    // Set a custom FVF for the mesh
    m_mesh->SetFVF(m_pd3dDevice, BLENDVERTEX_FVF);

    compute_weights();

    return S_OK;
}

//----------------------------------------------------------
// RestoreDeviceObjects()
//
// Restore the mesh, cached mesh data, and other device
// state.
//
HRESULT
CMyD3DApplication::RestoreDeviceObjects()
{
    // Restore mesh's local memory objects
    m_mesh->RestoreDeviceObjects(m_pd3dDevice);

    // Get access to the mesh vertex and index buffers
    {
        ID3DXMesh *mesh = m_mesh->GetLocalMesh();
        THR(mesh->GetVertexBuffer(&m_vertices));
        THR(mesh->GetIndexBuffer(&m_indices));
        m_num_vertices = mesh->GetNumVertices();
        m_num_faces = mesh->GetNumFaces();
    }

    if ((m_dwCreateFlags & D3DCREATE_SOFTWARE_VERTEXPROCESSING)
        || m_d3dCaps.VertexShaderVersion >= D3DVS_VERSION(1,0))
    {
        // Setup the vertex declaration
        DWORD adwDecl[50];
        THR(::D3DXDeclaratorFromFVF(BLENDVERTEX_FVF, adwDecl));

        // Create vertex shader from a file
        if (FAILED(::D3DUtil_CreateVertexShader(m_pd3dDevice, 
            _T("Blend.vsh"), adwDecl, &m_vertex_shader)))
        {
            return E_FAIL;
        }
    }
    // Set miscellaneous render states
    const rt::s_rs states[] =
    {
        D3DRS_ZENABLE, true,
        D3DRS_AMBIENT, 0x00404040,
        D3DRS_LIGHTING, true
    };
    rt::set_states(m_pd3dDevice, states, NUM_OF(states));


    // Set the projection matrix
    const float aspect = float(m_d3dsdBackBuffer.Width) /
        m_d3dsdBackBuffer.Height;
    ::D3DXMatrixPerspectiveFovLH(&m_projection, D3DX_PI/4,
        aspect, 1.0f, 10000.0f);
    THR(m_pd3dDevice->SetTransform(D3DTS_PROJECTION,
        &m_projection));

    // look_at(eye, look at, view up)
    m_view = rt::mat_look_at(D3DXVECTOR3(0.0f, -5.0f, -10.0f),
                             D3DXVECTOR3(0.0f, 0.0f,  0.0f),
                             D3DXVECTOR3(0.0f, 1.0f,  0.0f));
    THR(m_pd3dDevice->SetTransform(D3DTS_VIEW, &m_view));

    // Create two directional lights
    D3DLIGHT8 light;
    ::D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL,
        -0.5f, -1.0f, 1.0f);
    const float intensity = 0.9f;
    light.Diffuse.r = intensity;
    light.Diffuse.g = intensity;
    light.Diffuse.b = intensity;
    THR(m_pd3dDevice->SetLight(0, &light));
    THR(m_pd3dDevice->LightEnable(0, true));

    // Restore the font
    m_font->RestoreDeviceObjects();

    return S_OK;
}

//----------------------------------------------------------
// FrameMove()
//
// Change the world matrix over time; each transformation
// corresponds to a menu item.
//
HRESULT CMyD3DApplication::FrameMove()
{
    // Update user input state
    UpdateInput(&m_user_input);

    // Update the model rotation state
    {
        const bool left = m_user_input.bRotateLeft;
        const bool right = m_user_input.bRotateRight;
        const bool up = m_user_input.bRotateUp;
        const bool down = m_user_input.bRotateDown;
        if (left && !right)
        {
            m_rot_y += m_fElapsedTime;
        }
        else if (right && !left)
        {
            m_rot_y -= m_fElapsedTime;
        }
        if (up && !down)
        {
            m_rot_x += m_fElapsedTime;
        }
        else if (down && !up)
        {
            m_rot_x -= m_fElapsedTime;
        }
    }

    // Set the vertex blending matrices for this frame
    switch (m_trans)
    {
    case SKIN_TR_AXIS_ROTATION:
        {
            // rotate around an oscillating axis
            const D3DXVECTOR3 axis(2.f+sinf(3.1f*m_fTime),
                              2.f+sinf(3.3f*m_fTime),
                              sinf(3.5f*m_fTime));
            ::D3DXMatrixRotationAxis(&m_lower_arm,
                &axis, sinf(3*m_fTime));
        }
        break;

    case SKIN_TR_CENTER_SCALING:
        {
            // scale about mesh center
            const D3DXVECTOR3 center = 0.5f*
                (m_mesh_bounds.m_min + m_mesh_bounds.m_max);
            const float s = 0.75f + 0.5f*sinf(3*m_fTime);
            m_lower_arm = rt::mat_trans(-center)*
                rt::mat_scale(s)*
                rt::mat_trans(center);
        }
        break;

    case SKIN_TR_Y_TRANSLATE:
        // Y-axis translation
        ::D3DXMatrixTranslation(&m_lower_arm,
                0.0f, 3*sinf(3*m_fTime), 0.0f);
        break;

    case SKIN_TR_AXIS_ROTATION_LEFT:
        {
            // rotate around an oscillating axis about
            // the mesh's minimum
            const D3DXVECTOR3 axis(2.f + sinf(3.1f*m_fTime),
                                    2.f + sinf(3.3f*m_fTime),
                                    sinf(3.5f*m_fTime));
            m_lower_arm =
                rt::mat_trans(-m_mesh_bounds.m_min)*
                rt::mat_rot_axis(axis, sinf(3*m_fTime))*
                rt::mat_trans(m_mesh_bounds.m_min);
        }
        break;

    case SKIN_TR_Y_ROTATION_LEFT:
        {
            // Y-axis rotation around min_x of mesh
            m_lower_arm =
                rt::mat_trans(-m_mesh_bounds.m_min)*
                rt::mat_rot_y(D3DX_PI/4.0f +
                    D3DX_PI*sinf(3*m_fTime)/8.0f)*
                rt::mat_trans(m_mesh_bounds.m_min);
        }
        break;

    case SKIN_TR_Y_ROTATION_CENTER:
        {
            // Y-axis rotation around center of mesh
            const D3DXVECTOR3 center = 0.5f*
                (m_mesh_bounds.m_min + m_mesh_bounds.m_max);
            m_lower_arm = rt::mat_trans(center)*
                rt::mat_rot_y(D3DX_PI/4.0f +
                    D3DX_PI*sinf(3*m_fTime)/8.0f)*
                rt::mat_trans(center);
        }
        break;
    }
    ::D3DXMatrixIdentity(&m_upper_arm);

    // post-multiply orientation transformation onto
    // blend transformation
    const D3DXMATRIX rot = rt::mat_rot_x(m_rot_x)*
        rt::mat_rot_y(m_rot_y);
    m_upper_arm *= rot;
    m_lower_arm *= rot;

    // Set the vertex shader constants.
    if (m_use_vertex_shader)
    {
        // Some basic constants
        const D3DXVECTOR4 zero(0,0,0,0);
        const D3DXVECTOR4 one(1,1,1,1);

        // Lighting vector (normalized) and material colors.
        // (Use red light to show difference from non-vertex
        // shader case.)
        D3DXVECTOR4 light_dir(0.5f, 1.0f, -1.0f, 0.0f);
        ::D3DXVec4Normalize(&light_dir, &light_dir);
        const float diffuse[] = { 1.00f, 1.00f, 0.00f, 0.00f };
        const float ambient[] = { 0.25f, 0.25f, 0.25f, 0.25f };

        // compute transposed matrices used by the shader
        const D3DXMATRIX view_proj = m_view * m_projection;

        // Set the vertex shader constants; the shader uses
        // matrices with rows/columns reversed from D3DMATRIX,
        // so compute transpose of matrix before storing
#define SVSC(addr_, data_, num_) \
THR(m_pd3dDevice->SetVertexShaderConstant(addr_, data_, num_))
        SVSC(0, &zero, 1);
        SVSC(1, &one, 1);
        D3DXMATRIX transpose;
        ::D3DXMatrixTranspose(&transpose, &m_upper_arm);
        SVSC(4, &transpose, 4);
        ::D3DXMatrixTranspose(&transpose, &m_lower_arm);
        SVSC(8, &transpose, 4);
        ::D3DXMatrixTranspose(&transpose, &view_proj);
        SVSC(12, &transpose, 4);
        SVSC(20, &light_dir, 1);
        SVSC(21, diffuse, 1);
        SVSC(22, ambient, 1);
#undef SVSC
    }

    return S_OK;
}

//----------------------------------------------------------
// Render()
//
// Set either vertex shader or vertex blend state, then
// draw mesh.  Draw stats if requested.
//
HRESULT CMyD3DApplication::Render()
{
    // Clear the backbuffer
    THR(m_pd3dDevice->Clear(0L, NULL,
        D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 
        D3DCOLOR_XRGB(0, 0, 255), 1.0f, 0L));

    // Begin the scene
    THR(m_pd3dDevice->BeginScene());
    if (m_use_vertex_shader)        
    {
        THR(m_pd3dDevice->SetVertexShader(m_vertex_shader));
        THR(m_pd3dDevice->SetStreamSource(0, m_vertices,
            sizeof(BLENDVERTEX)));
        THR(m_pd3dDevice->SetIndices(m_indices, 0));
        THR(m_pd3dDevice->DrawIndexedPrimitive(
            D3DPT_TRIANGLELIST, 0, m_num_vertices,
            0, m_num_faces));
    }
    else
    {
        // Enable vertex blending using API
        THR(m_pd3dDevice->SetTransform(D3DTS_WORLD,
            &m_upper_arm));
        THR(m_pd3dDevice->SetTransform(D3DTS_WORLD1,
            &m_lower_arm));
        THR(m_pd3dDevice->SetRenderState(D3DRS_VERTEXBLEND,
            D3DVBF_1WEIGHTS));

        // Display the object
        m_mesh->Render(m_pd3dDevice);
    }

    // Output statistics
    if (m_show_stats)
    {
        RenderText();
    }
    THR(m_pd3dDevice->EndScene());

    return S_OK;
}

//----------------------------------------------------------
// compute_weights
//
// Add blending weights to the mesh based on m_weight_dist.
// Weights are distributed along the x-axis of the model.
//
void
CMyD3DApplication::compute_weights()
{
    // Gain acces to the mesh's vertices
    rt::mesh_vertex_lock<BLENDVERTEX>
        lock(m_mesh->GetSysMemMesh());
    BLENDVERTEX *verts = lock.data();

    // Set the blend factors for the vertices
    const UINT num_vertices =
        m_mesh->GetSysMemMesh()->GetNumVertices();
    for (UINT i = 0; i < num_vertices; i++)
    {
        // find fraction along x-axis extent for this vertex
        const float a =
            (verts[i].v.x - m_mesh_bounds.m_min.x) /
            (m_mesh_bounds.m_max.x - m_mesh_bounds.m_min.x);

        // apply weight distribution function
        switch (m_weight_dist)
        {
        case SKIN_WD_SINUSOIDAL:
            verts[i].blend = 1.0f - sinf(a*D3DX_PI);
            break;

        case SKIN_WD_TRIANGLE:
            verts[i].blend =
                a <= 0.5f ? 2.0f*a : 1.0f - 2*(a-0.5f);
            break;

        case SKIN_WD_BRACKETED:
            verts[i].blend =
                (a < 0.3f) ? 0.0f : ((a > 0.7f) ? 1.0f :
                (a-0.3f)/0.4f);
            break;

        case SKIN_WD_INVERTED_FRACTION:
            verts[i].blend = 1.0f - a;
            break;

        case SKIN_WD_FRACTION:
            verts[i].blend = a;
            break;

        case SKIN_WD_SIGMOIDAL:
            verts[i].blend =
                1.0f/(1.0f + expf(-25.0f*(a-0.5f)));
            break;
        }
    }
}
